@@ -28,21 +28,33 @@
  */
 #ifdef HAVE_CONFIG_H
 #include <config.h>
-#endif /* !HAVE_CONFIG_H */
+#endif /* HAVE_CONFIG_H */
 
-#include <stdint.h>
+#include <inttypes.h>
 #include <stdlib.h>
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif /* HAVE_UNISTD_H */
+#ifdef HAVE_FCNTL_H
 #include <fcntl.h>
+#endif /* HAVE_FCNTL_H */
 #include <sys/types.h>
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
+#endif /* HAVE_SYS_SOCKET_H */
+#ifdef HAVE_NETDB_H
 #include <netdb.h>
+#endif /* HAVE_NETDB_H */
+#ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
+#endif /* HAVE_NETINET_IN_H */
 #include <netinet/tcp.h>
 #include <poll.h>
 #include <signal.h>
 #include <stdio.h>
 #include <assert.h>
+#include <string.h>
+#include <errno.h>
 
 #include <nghttp2/nghttp2.h>
 
@@ -114,6 +126,7 @@ static char *strcopy(const char *s, size_t len) {
 /*
  * Prints error message |msg| and exit.
  */
+NGHTTP2_NORETURN
 static void die(const char *msg) {
   fprintf(stderr, "FATAL: %s\n", msg);
   exit(EXIT_FAILURE);
@@ -123,6 +136,7 @@ static void die(const char *msg) {
  * Prints error containing the function name |func| and message |msg|
  * and exit.
  */
+NGHTTP2_NORETURN
 static void dief(const char *func, const char *msg) {
   fprintf(stderr, "FATAL: %s: %s\n", func, msg);
   exit(EXIT_FAILURE);
@@ -132,6 +146,7 @@ static void dief(const char *func, const char *msg) {
  * Prints error containing the function name |func| and error code
  * |error_code| and exit.
  */
+NGHTTP2_NORETURN
 static void diec(const char *func, int error_code) {
   fprintf(stderr, "FATAL: %s: error_code=%d, msg=%s\n", func, error_code,
           nghttp2_strerror(error_code));
@@ -152,7 +167,7 @@ static ssize_t send_callback(nghttp2_session *session _U_, const uint8_t *data,
   connection->want_io = IO_NONE;
   ERR_clear_error();
   rv = SSL_write(connection->ssl, data, (int)length);
-  if (rv < 0) {
+  if (rv <= 0) {
     int err = SSL_get_error(connection->ssl, rv);
     if (err == SSL_ERROR_WANT_WRITE || err == SSL_ERROR_WANT_READ) {
       connection->want_io =
@@ -528,10 +543,6 @@ static void fetch_uri(const struct URI *uri) {
   connection.ssl = ssl;
   connection.want_io = IO_NONE;
 
-  /* Send connection header in blocking I/O mode */
-  SSL_write(ssl, NGHTTP2_CLIENT_CONNECTION_PREFACE,
-            NGHTTP2_CLIENT_CONNECTION_PREFACE_LEN);
-
   /* Here make file descriptor non-block */
   make_non_block(fd);
   set_tcp_nodelay(fd);
@@ -554,6 +565,8 @@ static void fetch_uri(const struct URI *uri) {
     diec("nghttp2_session_client_new", rv);
   }
 
+  nghttp2_submit_settings(connection.session, NGHTTP2_FLAG_NONE, NULL, 0);
+
   /* Submit the HTTP request to the outbound queue. */
   submit_request(&connection, &req);
 
@@ -649,10 +662,10 @@ static int parse_uri(struct URI *res, const char *uri) {
         return -1;
       }
       offset = i;
-      res->port = port;
+      res->port = (uint16_t)port;
     }
   }
-  res->hostportlen = uri + offset + ipv6addr - res->host;
+  res->hostportlen = (size_t)(uri + offset + ipv6addr - res->host);
   for (i = offset; i < len; ++i) {
     if (uri[i] == '#') {
       break;
@@ -681,8 +694,9 @@ int main(int argc, char **argv) {
   act.sa_handler = SIG_IGN;
   sigaction(SIGPIPE, &act, 0);
 
+#ifndef OPENSSL_IS_BORINGSSL
   OPENSSL_config(NULL);
-  OpenSSL_add_all_algorithms();
+#endif /* OPENSSL_IS_BORINGSSL */
   SSL_load_error_strings();
   SSL_library_init();
 
