@@ -29,11 +29,20 @@
 #include <config.h>
 #endif /* HAVE_CONFIG_H */
 
+#include <string.h>
+#include <stddef.h>
+
 #include <nghttp2/nghttp2.h>
+#include "nghttp2_mem.h"
 
 #define nghttp2_min(A, B) ((A) < (B) ? (A) : (B))
 #define nghttp2_max(A, B) ((A) > (B) ? (A) : (B))
 
+#define lstreq(A, B, N) ((sizeof((A)) - 1) == (N) && memcmp((A), (B), (N)) == 0)
+
+#define nghttp2_struct_of(ptr, type, member)                                   \
+  ((type *)(void *)((char *)(ptr) - offsetof(type, member)))
+
 /*
  * Copies 2 byte unsigned integer |n| in host byte order to |buf| in
  * network byte order.
@@ -58,18 +67,6 @@ uint16_t nghttp2_get_uint16(const uint8_t *data);
  */
 uint32_t nghttp2_get_uint32(const uint8_t *data);
 
-/*
- * Allocates |n| bytes of memory and copy the memory region pointed by
- * |src| with the length |n| bytes into it. Returns the allocated memory.
- *
- * This function returns pointer to allocated memory, or one of the
- * following negative error codes:
- *
- * NGHTTP2_ERR_NOMEM
- *     Out of memory.
- */
-void *nghttp2_memdup(const void *src, size_t n);
-
 void nghttp2_downcase(uint8_t *s, size_t len);
 
 /*
@@ -99,14 +96,6 @@ int nghttp2_adjust_local_window_size(int32_t *local_window_size_ptr,
 int nghttp2_should_send_window_update(int32_t local_window_size,
                                       int32_t recv_window_size);
 
-/*
- * Deallocates memory space pointed by |ptr|. This function exists for
- * the application to free the memory space allocated by the library
- * functions. Currently this function is hidden from the public API,
- * but may be exposed as public API.
- */
-void nghttp2_free(void *ptr);
-
 /*
  * Copies the buffer |src| of length |len| to the destination pointed
  * by the |dest|, assuming that the |dest| is at lest |len| bytes long
