@@ -35,11 +35,6 @@
 #include "nghttp2_pq.h"
 #include "nghttp2_int.h"
 
-/*
- * Maximum number of streams in one dependency tree.
- */
-#define NGHTTP2_MAX_DEP_TREE_LENGTH 100
-
 /*
  * If local peer is stream initiator:
  * NGHTTP2_STREAM_OPENING : upon sending request HEADERS
@@ -84,7 +79,8 @@ typedef enum {
 
 typedef enum {
   NGHTTP2_STREAM_FLAG_NONE = 0,
-  /* Indicates that this stream is pushed stream */
+  /* Indicates that this stream is pushed stream and not opened
+     yet. */
   NGHTTP2_STREAM_FLAG_PUSH = 0x01,
   /* Indicates that this stream was closed */
   NGHTTP2_STREAM_FLAG_CLOSED = 0x02,
@@ -98,24 +94,56 @@ typedef enum {
 
 } nghttp2_stream_flag;
 
+/* HTTP related flags to enforce HTTP semantics */
 typedef enum {
-  NGHTTP2_STREAM_DPRI_NONE = 0,
-  NGHTTP2_STREAM_DPRI_NO_DATA = 0x01,
-  NGHTTP2_STREAM_DPRI_TOP = 0x02,
-  NGHTTP2_STREAM_DPRI_REST = 0x04
-} nghttp2_stream_dpri;
-
-struct nghttp2_stream_roots;
-
-typedef struct nghttp2_stream_roots nghttp2_stream_roots;
-
-struct nghttp2_stream;
-
-typedef struct nghttp2_stream nghttp2_stream;
+  NGHTTP2_HTTP_FLAG_NONE = 0,
+  /* header field seen so far */
+  NGHTTP2_HTTP_FLAG__AUTHORITY = 1,
+  NGHTTP2_HTTP_FLAG__PATH = 1 << 1,
+  NGHTTP2_HTTP_FLAG__METHOD = 1 << 2,
+  NGHTTP2_HTTP_FLAG__SCHEME = 1 << 3,
+  /* host is not pseudo header, but we require either host or
+     :authority */
+  NGHTTP2_HTTP_FLAG_HOST = 1 << 4,
+  NGHTTP2_HTTP_FLAG__STATUS = 1 << 5,
+  /* required header fields for HTTP request except for CONNECT
+     method. */
+  NGHTTP2_HTTP_FLAG_REQ_HEADERS = NGHTTP2_HTTP_FLAG__METHOD |
+                                  NGHTTP2_HTTP_FLAG__PATH |
+                                  NGHTTP2_HTTP_FLAG__SCHEME,
+  NGHTTP2_HTTP_FLAG_PSEUDO_HEADER_DISALLOWED = 1 << 6,
+  /* HTTP method flags */
+  NGHTTP2_HTTP_FLAG_METH_CONNECT = 1 << 7,
+  NGHTTP2_HTTP_FLAG_METH_HEAD = 1 << 8,
+  NGHTTP2_HTTP_FLAG_METH_OPTIONS = 1 << 9,
+  NGHTTP2_HTTP_FLAG_METH_ALL = NGHTTP2_HTTP_FLAG_METH_CONNECT |
+                               NGHTTP2_HTTP_FLAG_METH_HEAD |
+                               NGHTTP2_HTTP_FLAG_METH_OPTIONS,
+  /* :path category */
+  /* path starts with "/" */
+  NGHTTP2_HTTP_FLAG_PATH_REGULAR = 1 << 10,
+  /* path "*" */
+  NGHTTP2_HTTP_FLAG_PATH_ASTERISK = 1 << 11,
+  /* scheme */
+  /* "http" or "https" scheme */
+  NGHTTP2_HTTP_FLAG_SCHEME_HTTP = 1 << 12,
+  /* set if final response is expected */
+  NGHTTP2_HTTP_FLAG_EXPECT_FINAL_RESPONSE = 1 << 13
+} nghttp2_http_flag;
 
 struct nghttp2_stream {
   /* Intrusive Map */
   nghttp2_map_entry map_entry;
+  /* Entry for dep_prev->obq */
+  nghttp2_pq_entry pq_entry;
+  /* Priority Queue storing direct descendant (nghttp2_stream).  Only
+     streams which itself has some data to send, or has a descendant
+     which has some data to sent. */
+  nghttp2_pq obq;
+  /* Content-Length of request/response body.  -1 if unknown. */
+  int64_t content_length;
+  /* Received body so far */
+  int64_t recv_content_length;
   /* pointers to form dependency tree.  If multiple streams depend on
      a stream, only one stream (left most) has non-NULL dep_prev which
      points to the stream it depends on. The remaining streams are
@@ -125,28 +153,17 @@ struct nghttp2_stream {
      dep_prev and sib_prev are NULL. */
   nghttp2_stream *dep_prev, *dep_next;
   nghttp2_stream *sib_prev, *sib_next;
-  /* pointers to track dependency tree root streams.  This is
-     doubly-linked list and first element is pointed by
-     roots->head. */
-  nghttp2_stream *root_prev, *root_next;
   /* When stream is kept after closure, it may be kept in doubly
      linked list pointed by nghttp2_session closed_stream_head.
      closed_next points to the next stream object if it is the element
      of the list. */
   nghttp2_stream *closed_prev, *closed_next;
-  /* pointer to roots, which tracks dependency tree roots */
-  nghttp2_stream_roots *roots;
   /* The arbitrary data provided by user for this stream. */
   void *stream_user_data;
   /* Item to send */
-  nghttp2_outbound_item *data_item;
+  nghttp2_outbound_item *item;
   /* stream ID */
   int32_t stream_id;
-  /* categorized priority of this stream.  Only stream bearing
-     NGHTTP2_STREAM_DPRI_TOP can send item. */
-  nghttp2_stream_dpri dpri;
-  /* the number of streams in subtree */
-  size_t num_substreams;
   /* Current remote window size. This value is computed against the
      current initial window size of remote endpoint. */
   int32_t remote_window_size;
@@ -167,30 +184,39 @@ struct nghttp2_stream {
   int32_t local_window_size;
   /* weight of this stream */
   int32_t weight;
-  /* effective weight of this stream in belonging dependency tree */
-  int32_t effective_weight;
-  /* sum of weight (not effective_weight) of direct descendants */
+  /* sum of weight of direct descendants */
   int32_t sum_dep_weight;
-  /* sum of weight of direct descendants which have at least one
-     descendant with dpri == NGHTTP2_STREAM_DPRI_TOP.  We use this
-     value to calculate effective weight. */
-  int32_t sum_norest_weight;
-  /* sum of weight of direct descendants whose dpri value is
-     NGHTTP2_STREAM_DPRI_TOP */
-  int32_t sum_top_weight;
   nghttp2_stream_state state;
+  /* status code from remote server */
+  int16_t status_code;
+  /* Bitwise OR of zero or more nghttp2_http_flag values */
+  uint16_t http_flags;
   /* This is bitwise-OR of 0 or more of nghttp2_stream_flag. */
   uint8_t flags;
   /* Bitwise OR of zero or more nghttp2_shut_flag values */
   uint8_t shut_flags;
+  /* Nonzero if this stream has been queued to stream pointed by
+     dep_prev.  We maintain the invariant that if a stream is queued,
+     then its ancestors, except for root, are also queued.  This
+     invariant may break in fatal error condition. */
+  uint8_t queued;
+  /* Base last_cycle for direct descendent streams. */
+  uint64_t descendant_last_cycle;
+  /* Next scheduled time to sent item */
+  uint64_t cycle;
+  /* Last written length of frame payload */
+  size_t last_writelen;
+  /* This flag is used to reduce excessive queuing of WINDOW_UPDATE to
+     this stream.  The nonzero does not necessarily mean WINDOW_UPDATE
+     is not queued. */
+  uint8_t window_update_queued;
 };
 
 void nghttp2_stream_init(nghttp2_stream *stream, int32_t stream_id,
                          uint8_t flags, nghttp2_stream_state initial_state,
-                         int32_t weight, nghttp2_stream_roots *roots,
-                         int32_t remote_initial_window_size,
+                         int32_t weight, int32_t remote_initial_window_size,
                          int32_t local_initial_window_size,
-                         void *stream_user_data);
+                         void *stream_user_data, nghttp2_mem *mem);
 
 void nghttp2_stream_free(nghttp2_stream *stream);
 
@@ -201,9 +227,9 @@ void nghttp2_stream_free(nghttp2_stream *stream);
 void nghttp2_stream_shutdown(nghttp2_stream *stream, nghttp2_shut_flag flag);
 
 /*
- * Defer |stream->data_item|.  We won't call this function in the
- * situation where |stream->data_item| == NULL.  If |flags| is bitwise
- * OR of zero or more of NGHTTP2_STREAM_FLAG_DEFERRED_USER and
+ * Defer |stream->item|.  We won't call this function in the situation
+ * where |stream->item| == NULL.  The |flags| is bitwise OR of zero or
+ * more of NGHTTP2_STREAM_FLAG_DEFERRED_USER and
  * NGHTTP2_STREAM_FLAG_DEFERRED_FLOW_CONTROL.  The |flags| indicates
  * the reason of this action.
  *
@@ -213,8 +239,7 @@ void nghttp2_stream_shutdown(nghttp2_stream *stream, nghttp2_shut_flag flag);
  * NGHTTP2_ERR_NOMEM
  *     Out of memory
  */
-int nghttp2_stream_defer_data(nghttp2_stream *stream, uint8_t flags,
-                              nghttp2_session *session);
+int nghttp2_stream_defer_item(nghttp2_stream *stream, uint8_t flags);
 
 /*
  * Put back deferred data in this stream to active state.  The |flags|
@@ -223,17 +248,22 @@ int nghttp2_stream_defer_data(nghttp2_stream *stream, uint8_t flags,
  * NGHTTP2_STREAM_FLAG_DEFERRED_FLOW_CONTROL and given masks are
  * cleared if they are set.  So even if this function is called, if
  * one of flag is still set, data does not become active.
+ *
+ * This function returns 0 if it succeeds, or one of the following
+ * negative error codes:
+ *
+ * NGHTTP2_ERR_NOMEM
+ *     Out of memory
  */
-int nghttp2_stream_resume_deferred_data(nghttp2_stream *stream, uint8_t flags,
-                                        nghttp2_session *session);
+int nghttp2_stream_resume_deferred_item(nghttp2_stream *stream, uint8_t flags);
 
 /*
- * Returns nonzero if data item is deferred by whatever reason.
+ * Returns nonzero if item is deferred by whatever reason.
  */
-int nghttp2_stream_check_deferred_data(nghttp2_stream *stream);
+int nghttp2_stream_check_deferred_item(nghttp2_stream *stream);
 
 /*
- * Returns nonzero if data item is deferred by flow control.
+ * Returns nonzero if item is deferred by flow control.
  */
 int nghttp2_stream_check_deferred_by_flow_control(nghttp2_stream *stream);
 
@@ -269,61 +299,47 @@ int nghttp2_stream_update_local_initial_window_size(
 void nghttp2_stream_promise_fulfilled(nghttp2_stream *stream);
 
 /*
- * Returns the stream positioned in root of the dependency tree the
- * |stream| belongs to.
+ * Returns nonzero if |target| is an ancestor of |stream|.
  */
-nghttp2_stream *nghttp2_stream_get_dep_root(nghttp2_stream *stream);
-
-/*
- * Returns nonzero if |target| is found in subtree of |stream|.
- */
-int nghttp2_stream_dep_subtree_find(nghttp2_stream *stream,
-                                    nghttp2_stream *target);
+int nghttp2_stream_dep_find_ancestor(nghttp2_stream *stream,
+                                     nghttp2_stream *target);
 
 /*
  * Computes distributed weight of a stream of the |weight| under the
- * |stream| if |stream| is removed from a dependency tree.  The result
- * is computed using stream->weight rather than
- * stream->effective_weight.
+ * |stream| if |stream| is removed from a dependency tree.
  */
 int32_t nghttp2_stream_dep_distributed_weight(nghttp2_stream *stream,
                                               int32_t weight);
 
-/*
- * Computes effective weight of a stream of the |weight| under the
- * |stream|.  The result is computed using stream->effective_weight
- * rather than stream->weight.  This function is used to determine
- * weight in dependency tree.
- */
-int32_t nghttp2_stream_dep_distributed_effective_weight(nghttp2_stream *stream,
-                                                        int32_t weight);
-
 /*
  * Makes the |stream| depend on the |dep_stream|.  This dependency is
  * exclusive.  All existing direct descendants of |dep_stream| become
  * the descendants of the |stream|.  This function assumes
- * |stream->data| is NULL and no dpri members are changed in this
- * dependency tree.
+ * |stream->item| is NULL.
+ *
+ * This function returns 0 if it succeeds, or one of the following
+ * negative error codes:
+ *
+ * NGHTTP2_ERR_NOMEM
+ *     Out of memory
  */
-void nghttp2_stream_dep_insert(nghttp2_stream *dep_stream,
-                               nghttp2_stream *stream);
+int nghttp2_stream_dep_insert(nghttp2_stream *dep_stream,
+                              nghttp2_stream *stream);
 
 /*
  * Makes the |stream| depend on the |dep_stream|.  This dependency is
- * not exclusive.  This function assumes |stream->data| is NULL and no
- * dpri members are changed in this dependency tree.
+ * not exclusive.  This function assumes |stream->item| is NULL.
  */
 void nghttp2_stream_dep_add(nghttp2_stream *dep_stream, nghttp2_stream *stream);
 
 /*
  * Removes the |stream| from the current dependency tree.  This
- * function assumes |stream->data| is NULL.
+ * function assumes |stream->item| is NULL.
  */
-void nghttp2_stream_dep_remove(nghttp2_stream *stream);
+int nghttp2_stream_dep_remove(nghttp2_stream *stream);
 
 /*
- * Attaches |data_item| to |stream|.  Updates dpri members in this
- * dependency tree.
+ * Attaches |item| to |stream|.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -331,14 +347,12 @@ void nghttp2_stream_dep_remove(nghttp2_stream *stream);
  * NGHTTP2_ERR_NOMEM
  *     Out of memory
  */
-int nghttp2_stream_attach_data(nghttp2_stream *stream,
-                               nghttp2_outbound_item *data_item,
-                               nghttp2_session *session);
+int nghttp2_stream_attach_item(nghttp2_stream *stream,
+                               nghttp2_outbound_item *item);
 
 /*
- * Detaches |stream->data_item|.  Updates dpri members in this
- * dependency tree.  This function does not free |stream->data_item|.
- * The caller must free it.
+ * Detaches |stream->item|.  This function does not free
+ * |stream->item|.  The caller must free it.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -346,12 +360,11 @@ int nghttp2_stream_attach_data(nghttp2_stream *stream,
  * NGHTTP2_ERR_NOMEM
  *     Out of memory
  */
-int nghttp2_stream_detach_data(nghttp2_stream *stream,
-                               nghttp2_session *session);
+int nghttp2_stream_detach_item(nghttp2_stream *stream);
 
 /*
  * Makes the |stream| depend on the |dep_stream|.  This dependency is
- * exclusive.  Updates dpri members in this dependency tree.
+ * exclusive.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -360,12 +373,11 @@ int nghttp2_stream_detach_data(nghttp2_stream *stream,
  *     Out of memory
  */
 int nghttp2_stream_dep_insert_subtree(nghttp2_stream *dep_stream,
-                                      nghttp2_stream *stream,
-                                      nghttp2_session *session);
+                                      nghttp2_stream *stream);
 
 /*
  * Makes the |stream| depend on the |dep_stream|.  This dependency is
- * not exclusive.  Updates dpri members in this dependency tree.
+ * not exclusive.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -374,13 +386,11 @@ int nghttp2_stream_dep_insert_subtree(nghttp2_stream *dep_stream,
  *     Out of memory
  */
 int nghttp2_stream_dep_add_subtree(nghttp2_stream *dep_stream,
-                                   nghttp2_stream *stream,
-                                   nghttp2_session *session);
+                                   nghttp2_stream *stream);
 
 /*
- * Removes subtree whose root stream is |stream|.  Removing subtree
- * does not change dpri values.  The effective_weight of streams in
- * removed subtree is not updated.
+ * Removes subtree whose root stream is |stream|.  The
+ * effective_weight of streams in removed subtree is not updated.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -390,54 +400,21 @@ int nghttp2_stream_dep_add_subtree(nghttp2_stream *dep_stream,
  */
 void nghttp2_stream_dep_remove_subtree(nghttp2_stream *stream);
 
-/*
- * Makes the |stream| as root.  Updates dpri members in this
- * dependency tree.
- *
- * This function returns 0 if it succeeds, or one of the following
- * negative error codes:
- *
- * NGHTTP2_ERR_NOMEM
- *     Out of memory
- */
-int nghttp2_stream_dep_make_root(nghttp2_stream *stream,
-                                 nghttp2_session *session);
-
-/*
- * Makes the |stream| as root and all existing root streams become
- * direct children of |stream|.
- *
- * This function returns 0 if it succeeds, or one of the following
- * negative error codes:
- *
- * NGHTTP2_ERR_NOMEM
- *     Out of memory
- */
-int
-nghttp2_stream_dep_all_your_stream_are_belong_to_us(nghttp2_stream *stream,
-                                                    nghttp2_session *session);
-
 /*
  * Returns nonzero if |stream| is in any dependency tree.
  */
 int nghttp2_stream_in_dep_tree(nghttp2_stream *stream);
 
-struct nghttp2_stream_roots {
-  nghttp2_stream *head;
+/*
+ * Schedules transmission of |stream|'s item, assuming stream->item is
+ * attached, and stream->last_writelen was updated.
+ */
+void nghttp2_stream_reschedule(nghttp2_stream *stream);
 
-  int32_t num_streams;
-};
-
-void nghttp2_stream_roots_init(nghttp2_stream_roots *roots);
-
-void nghttp2_stream_roots_free(nghttp2_stream_roots *roots);
-
-void nghttp2_stream_roots_add(nghttp2_stream_roots *roots,
-                              nghttp2_stream *stream);
-
-void nghttp2_stream_roots_remove(nghttp2_stream_roots *roots,
-                                 nghttp2_stream *stream);
-
-void nghttp2_stream_roots_remove_all(nghttp2_stream_roots *roots);
+/*
+ * Returns a stream which has highest priority.
+ */
+nghttp2_outbound_item *
+nghttp2_stream_next_outbound_item(nghttp2_stream *stream);
 
 #endif /* NGHTTP2_STREAM */
