@@ -34,21 +34,19 @@
 // file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 //
 
-#ifndef HTTP_SERVER2_SERVER_HPP
-#define HTTP_SERVER2_SERVER_HPP
+#ifndef ASIO_SERVER_H
+#define ASIO_SERVER_H
 
 #include "nghttp2_config.h"
 
 #include <string>
 #include <vector>
 #include <memory>
+
 #include <boost/noncopyable.hpp>
-#include <boost/asio.hpp>
-#include <boost/asio/ssl.hpp>
 
-#include <nghttp2/asio_http2.h>
+#include <nghttp2/asio_http2_server.h>
 
-#include "asio_connection.h"
 #include "asio_io_service_pool.h"
 
 namespace nghttp2 {
@@ -57,40 +55,45 @@ namespace asio_http2 {
 
 namespace server {
 
-/// The top-level class of the HTTP server.
+class serve_mux;
+
+using boost::asio::ip::tcp;
+
+using ssl_socket = boost::asio::ssl::stream<tcp::socket>;
+
 class server : private boost::noncopyable {
 public:
-  /// Construct the server to listen on the specified TCP address and port, and
-  /// serve up files from the given directory.
-  explicit server(const std::string &address, uint16_t port,
-                  std::size_t io_service_pool_size,
-                  std::size_t thread_pool_size, request_cb cb,
-                  std::unique_ptr<boost::asio::ssl::context> ssl_ctx,
-                  int backlog = -1);
+  explicit server(std::size_t io_service_pool_size);
 
-  /// Run the server's io_service loop.
-  void run();
+  boost::system::error_code
+  listen_and_serve(boost::system::error_code &ec,
+                   boost::asio::ssl::context *tls_context,
+                   const std::string &address, const std::string &port,
+                   int backlog, serve_mux &mux, bool asynchronous = false);
+  void join();
+  void stop();
 
 private:
   /// Initiate an asynchronous accept operation.
-  void start_accept();
+  void start_accept(tcp::acceptor &acceptor, serve_mux &mux);
+  /// Same as above but with tls_context
+  void start_accept(boost::asio::ssl::context &tls_context,
+                    tcp::acceptor &acceptor, serve_mux &mux);
 
-  void start_timer();
+  /// Resolves address and bind socket to the resolved addresses.
+  boost::system::error_code bind_and_listen(boost::system::error_code &ec,
+                                            const std::string &address,
+                                            const std::string &port,
+                                            int backlog);
 
-  /// The pool of io_service objects used to perform asynchronous operations.
+  /// The pool of io_service objects used to perform asynchronous
+  /// operations.
   io_service_pool io_service_pool_;
 
-  /// The signal_set is used to register for process termination notifications.
-  boost::asio::signal_set signals_;
-
-  boost::asio::deadline_timer tick_timer_;
-
   /// Acceptor used to listen for incoming connections.
-  std::vector<boost::asio::ip::tcp::acceptor> acceptors_;
+  std::vector<tcp::acceptor> acceptors_;
 
   std::unique_ptr<boost::asio::ssl::context> ssl_ctx_;
-
-  request_cb request_cb_;
 };
 
 } // namespace server
@@ -99,4 +102,4 @@ private:
 
 } // namespace nghttp2
 
-#endif // HTTP_SERVER2_SERVER_HPP
+#endif // ASIO_SERVER_H
