@@ -24,11 +24,9 @@
  */
 #include "shrpx_http2_downstream_connection.h"
 
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-
-#include <openssl/err.h>
-
-#include <event2/bufferevent_ssl.h>
+#endif // HAVE_UNISTD_H
 
 #include "http-parser/http_parser.h"
 
@@ -39,7 +37,6 @@
 #include "shrpx_error.h"
 #include "shrpx_http.h"
 #include "shrpx_http2_session.h"
-#include "shrpx_worker_config.h"
 #include "http2.h"
 #include "util.h"
 
@@ -49,16 +46,13 @@ namespace shrpx {
 
 Http2DownstreamConnection::Http2DownstreamConnection(
     DownstreamConnectionPool *dconn_pool, Http2Session *http2session)
-    : DownstreamConnection(dconn_pool), http2session_(http2session),
-      request_body_buf_(nullptr), sd_(nullptr) {}
+    : DownstreamConnection(dconn_pool), dlnext(nullptr), dlprev(nullptr),
+      http2session_(http2session), sd_(nullptr) {}
 
 Http2DownstreamConnection::~Http2DownstreamConnection() {
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Deleting";
   }
-  if (request_body_buf_) {
-    evbuffer_free(request_body_buf_);
-  }
   if (downstream_) {
     downstream_->disable_downstream_rtimer();
     downstream_->disable_downstream_wtimer();
@@ -73,69 +67,42 @@ Http2DownstreamConnection::~Http2DownstreamConnection() {
       error_code = NGHTTP2_INTERNAL_ERROR;
     }
 
-    if (LOG_ENABLED(INFO)) {
-      DCLOG(INFO, this) << "Submit RST_STREAM for DOWNSTREAM:" << downstream_
-                        << ", stream_id="
-                        << downstream_->get_downstream_stream_id()
-                        << ", error_code=" << error_code;
-    }
-
-    if (submit_rst_stream(downstream_, error_code) == 0) {
-      http2session_->notify();
-    }
-
     if (downstream_->get_downstream_stream_id() != -1) {
+      if (LOG_ENABLED(INFO)) {
+        DCLOG(INFO, this) << "Submit RST_STREAM for DOWNSTREAM:" << downstream_
+                          << ", stream_id="
+                          << downstream_->get_downstream_stream_id()
+                          << ", error_code=" << error_code;
+      }
+
+      submit_rst_stream(downstream_, error_code);
+
       http2session_->consume(downstream_->get_downstream_stream_id(),
                              downstream_->get_response_datalen());
 
       downstream_->reset_response_datalen();
 
-      http2session_->notify();
+      http2session_->signal_write();
     }
   }
   http2session_->remove_downstream_connection(this);
-  // Downstream and DownstreamConnection may be deleted
-  // asynchronously.
-  if (downstream_) {
-    downstream_->release_downstream_connection();
-  }
+
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Deleted";
   }
 }
 
-int Http2DownstreamConnection::init_request_body_buf() {
-  int rv;
-  if (request_body_buf_) {
-    rv = evbuffer_drain(request_body_buf_,
-                        evbuffer_get_length(request_body_buf_));
-    if (rv != 0) {
-      return -1;
-    }
-  } else {
-    request_body_buf_ = evbuffer_new();
-    if (request_body_buf_ == nullptr) {
-      return -1;
-    }
-  }
-  return 0;
-}
-
 int Http2DownstreamConnection::attach_downstream(Downstream *downstream) {
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Attaching to DOWNSTREAM:" << downstream;
   }
-  if (init_request_body_buf() == -1) {
-    return -1;
-  }
   http2session_->add_downstream_connection(this);
   if (http2session_->get_state() == Http2Session::DISCONNECTED) {
-    http2session_->notify();
+    http2session_->signal_write();
   }
 
   downstream_ = downstream;
-
-  downstream_->init_downstream_timer();
+  downstream_->reset_downstream_rtimer();
 
   return 0;
 }
@@ -145,7 +112,7 @@ void Http2DownstreamConnection::detach_downstream(Downstream *downstream) {
     DCLOG(INFO, this) << "Detaching from DOWNSTREAM:" << downstream;
   }
   if (submit_rst_stream(downstream) == 0) {
-    http2session_->notify();
+    http2session_->signal_write();
   }
 
   if (downstream_->get_downstream_stream_id() != -1) {
@@ -154,7 +121,7 @@ void Http2DownstreamConnection::detach_downstream(Downstream *downstream) {
 
     downstream_->reset_response_datalen();
 
-    http2session_->notify();
+    http2session_->signal_write();
   }
 
   downstream->disable_downstream_rtimer();
@@ -169,6 +136,7 @@ int Http2DownstreamConnection::submit_rst_stream(Downstream *downstream,
       downstream->get_downstream_stream_id() != -1) {
     switch (downstream->get_response_state()) {
     case Downstream::MSG_RESET:
+    case Downstream::MSG_BAD_HEADER:
     case Downstream::MSG_COMPLETE:
       break;
     default:
@@ -189,6 +157,7 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
                                  uint8_t *buf, size_t length,
                                  uint32_t *data_flags,
                                  nghttp2_data_source *source, void *user_data) {
+  int rv;
   auto sd = static_cast<StreamData *>(
       nghttp2_session_get_stream_user_data(session, stream_id));
   if (!sd || !sd->dconn) {
@@ -201,13 +170,9 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
     // on the priority, DATA frame may come first.
     return NGHTTP2_ERR_DEFERRED;
   }
-  auto body = dconn->get_request_body_buf();
-
-  auto nread = evbuffer_remove(body, buf, length);
-  if (nread == -1) {
-    DCLOG(FATAL, dconn) << "evbuffer_remove() failed";
-    return NGHTTP2_ERR_CALLBACK_FAILURE;
-  }
+  auto input = downstream->get_request_buf();
+  auto nread = input->remove(buf, length);
+  auto input_empty = input->rleft() == 0;
 
   if (nread > 0) {
     // This is important because it will handle flow control
@@ -225,7 +190,7 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
     }
   }
 
-  if (evbuffer_get_length(body) == 0 &&
+  if (input_empty &&
       downstream->get_request_state() == Downstream::MSG_COMPLETE &&
       // If connection is upgraded, don't set EOF flag, since HTTP/1
       // will set MSG_COMPLETE to request state after upgrade response
@@ -235,9 +200,26 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
         !downstream->get_upgraded()))) {
 
     *data_flags |= NGHTTP2_DATA_FLAG_EOF;
+
+    auto &trailers = downstream->get_request_trailers();
+    if (!trailers.empty()) {
+      std::vector<nghttp2_nv> nva;
+      nva.reserve(trailers.size());
+      http2::copy_headers_to_nva(nva, trailers);
+      if (!nva.empty()) {
+        rv = nghttp2_submit_trailer(session, stream_id, nva.data(), nva.size());
+        if (rv != 0) {
+          if (nghttp2_is_fatal(rv)) {
+            return NGHTTP2_ERR_CALLBACK_FAILURE;
+          }
+        } else {
+          *data_flags |= NGHTTP2_DATA_FLAG_NO_END_STREAM;
+        }
+      }
+    }
   }
 
-  if (evbuffer_get_length(body) > 0) {
+  if (!input_empty) {
     downstream->reset_downstream_wtimer();
   } else {
     downstream->disable_downstream_wtimer();
@@ -255,169 +237,129 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
 
 int Http2DownstreamConnection::push_request_headers() {
   int rv;
-  if (http2session_->get_state() != Http2Session::CONNECTED) {
-    // The HTTP2 session to the backend has not been established.
-    // This function will be called again just after it is
-    // established.
-    return 0;
-  }
   if (!downstream_) {
     return 0;
   }
-  size_t nheader = downstream_->get_request_headers().size();
-  if (!get_config()->http2_no_cookie_crumbling) {
-    downstream_->crumble_request_cookie();
+  if (!http2session_->can_push_request()) {
+    // The HTTP2 session to the backend has not been established or
+    // connection is now being checked.  This function will be called
+    // again just after it is established.
+    downstream_->set_request_pending(true);
+    http2session_->start_checking_connection();
+    return 0;
   }
 
-  assert(downstream_->get_request_headers_normalized());
+  downstream_->set_request_pending(false);
 
-  auto end_headers = std::end(downstream_->get_request_headers());
+  auto method = downstream_->get_request_method();
+  auto no_host_rewrite = get_config()->no_host_rewrite ||
+                         get_config()->http2_proxy ||
+                         get_config()->client_proxy || method == HTTP_CONNECT;
 
-  // 7 means:
+  // http2session_ has already in CONNECTED state, so we can get
+  // addr_idx here.
+  auto addr_idx = http2session_->get_addr_idx();
+  auto group = http2session_->get_group();
+  auto downstream_hostport = get_config()
+                                 ->downstream_addr_groups[group]
+                                 .addrs[addr_idx]
+                                 .hostport.get();
+
+  // For HTTP/1.0 request, there is no authority in request.  In that
+  // case, we use backend server's host nonetheless.
+  const char *authority = downstream_hostport;
+  auto &req_authority = downstream_->get_request_http2_authority();
+  if (no_host_rewrite && !req_authority.empty()) {
+    authority = req_authority.c_str();
+  }
+
+  if (!authority) {
+    authority = downstream_hostport;
+  }
+
+  downstream_->set_request_downstream_host(authority);
+
+  auto nheader = downstream_->get_request_headers().size();
+
+  Headers cookies;
+  if (!get_config()->http2_no_cookie_crumbling) {
+    cookies = downstream_->crumble_request_cookie();
+  }
+
+  // 8 means:
   // 1. :method
   // 2. :scheme
   // 3. :path
-  // 4. :authority (optional)
+  // 4. :authority
   // 5. via (optional)
   // 6. x-forwarded-for (optional)
   // 7. x-forwarded-proto (optional)
+  // 8. te (optional)
   auto nva = std::vector<nghttp2_nv>();
-  nva.reserve(nheader + 7);
-  std::string via_value;
-  std::string xff_value;
-  std::string scheme, authority, path, query;
-  // To reconstruct HTTP/1 status line and headers, proxy should
-  // preserve host header field. See draft-09 section 8.1.3.1.
-  if (downstream_->get_request_method() == "CONNECT") {
-    // The upstream may be HTTP/2 or HTTP/1
-    if (!downstream_->get_request_http2_authority().empty()) {
-      nva.push_back(http2::make_nv_ls(
-          ":authority", downstream_->get_request_http2_authority()));
-    } else {
-      nva.push_back(
-          http2::make_nv_ls(":authority", downstream_->get_request_path()));
-    }
-  } else if (!downstream_->get_request_http2_scheme().empty()) {
-    // Here the upstream is HTTP/2
-    nva.push_back(
-        http2::make_nv_ls(":scheme", downstream_->get_request_http2_scheme()));
-    nva.push_back(http2::make_nv_ls(":path", downstream_->get_request_path()));
-    if (!downstream_->get_request_http2_authority().empty()) {
-      nva.push_back(http2::make_nv_ls(
-          ":authority", downstream_->get_request_http2_authority()));
-    } else if (downstream_->get_norm_request_header("host") == end_headers) {
-      if (LOG_ENABLED(INFO)) {
-        DCLOG(INFO, this) << "host header field missing";
-      }
-      return -1;
-    }
-  } else {
-    // The upstream is HTTP/1
-    http_parser_url u;
-    const char *url = downstream_->get_request_path().c_str();
-    memset(&u, 0, sizeof(u));
-    rv = http_parser_parse_url(url, downstream_->get_request_path().size(), 0,
-                               &u);
-    if (rv == 0) {
-      http2::copy_url_component(scheme, &u, UF_SCHEMA, url);
-      http2::copy_url_component(authority, &u, UF_HOST, url);
-      http2::copy_url_component(path, &u, UF_PATH, url);
-      http2::copy_url_component(query, &u, UF_QUERY, url);
-      if (path.empty()) {
-        if (!authority.empty() &&
-            downstream_->get_request_method() == "OPTIONS") {
-          path = "*";
-        } else {
-          path = "/";
-        }
-      }
-      if (!query.empty()) {
-        path += "?";
-        path += query;
-      }
-    }
-    if (scheme.empty()) {
-      if (client_handler_->get_ssl()) {
-        nva.push_back(http2::make_nv_ll(":scheme", "https"));
-      } else {
-        nva.push_back(http2::make_nv_ll(":scheme", "http"));
-      }
-    } else {
-      nva.push_back(http2::make_nv_ls(":scheme", scheme));
-    }
-    if (path.empty()) {
-      nva.push_back(
-          http2::make_nv_ls(":path", downstream_->get_request_path()));
+  nva.reserve(nheader + 8 + cookies.size() +
+              get_config()->add_request_headers.size());
+
+  nva.push_back(http2::make_nv_lc(":method", http2::to_method_string(method)));
+
+  auto &scheme = downstream_->get_request_http2_scheme();
+
+  nva.push_back(http2::make_nv_lc(":authority", authority));
+
+  if (method != HTTP_CONNECT) {
+    assert(!scheme.empty());
+
+    nva.push_back(http2::make_nv_ls(":scheme", scheme));
+
+    auto &path = downstream_->get_request_path();
+    if (method == HTTP_OPTIONS && path.empty()) {
+      nva.push_back(http2::make_nv_ll(":path", "*"));
     } else {
       nva.push_back(http2::make_nv_ls(":path", path));
     }
-    if (!authority.empty()) {
-      // TODO properly check IPv6 numeric address
-      if (authority.find(":") != std::string::npos) {
-        authority = "[" + authority;
-        authority += "]";
-      }
-      if (u.field_set & (1 << UF_PORT)) {
-        authority += ":";
-        authority += util::utos(u.port);
-      }
-      nva.push_back(http2::make_nv_ls(":authority", authority));
-    } else if (downstream_->get_norm_request_header("host") == end_headers) {
-      if (LOG_ENABLED(INFO)) {
-        DCLOG(INFO, this) << "host header field missing";
-      }
-      return -1;
-    }
   }
 
-  nva.push_back(
-      http2::make_nv_ls(":method", downstream_->get_request_method()));
-
-  http2::copy_norm_headers_to_nva(nva, downstream_->get_request_headers());
+  http2::copy_headers_to_nva(nva, downstream_->get_request_headers());
 
   bool chunked_encoding = false;
   auto transfer_encoding =
-      downstream_->get_norm_request_header("transfer-encoding");
-  if (transfer_encoding != end_headers &&
-      util::strieq((*transfer_encoding).value.c_str(), "chunked")) {
+      downstream_->get_request_header(http2::HD_TRANSFER_ENCODING);
+  if (transfer_encoding &&
+      util::strieq_l("chunked", (*transfer_encoding).value)) {
     chunked_encoding = true;
   }
 
-  auto xff = downstream_->get_norm_request_header("x-forwarded-for");
+  for (auto &nv : cookies) {
+    nva.push_back(http2::make_nv(nv.name, nv.value, nv.no_index));
+  }
+
+  std::string xff_value;
+  auto xff = downstream_->get_request_header(http2::HD_X_FORWARDED_FOR);
   if (get_config()->add_x_forwarded_for) {
-    if (xff != end_headers && !get_config()->strip_incoming_x_forwarded_for) {
+    if (xff && !get_config()->strip_incoming_x_forwarded_for) {
       xff_value = (*xff).value;
       xff_value += ", ";
     }
     xff_value +=
         downstream_->get_upstream()->get_client_handler()->get_ipaddr();
     nva.push_back(http2::make_nv_ls("x-forwarded-for", xff_value));
-  } else if (xff != end_headers &&
-             !get_config()->strip_incoming_x_forwarded_for) {
+  } else if (xff && !get_config()->strip_incoming_x_forwarded_for) {
     nva.push_back(http2::make_nv_ls("x-forwarded-for", (*xff).value));
   }
 
   if (!get_config()->http2_proxy && !get_config()->client_proxy &&
-      downstream_->get_request_method() != "CONNECT") {
+      downstream_->get_request_method() != HTTP_CONNECT) {
     // We use same protocol with :scheme header field
-    if (scheme.empty()) {
-      if (client_handler_->get_ssl()) {
-        nva.push_back(http2::make_nv_ll("x-forwarded-proto", "https"));
-      } else {
-        nva.push_back(http2::make_nv_ll("x-forwarded-proto", "http"));
-      }
-    } else {
-      nva.push_back(http2::make_nv_ls("x-forwarded-proto", scheme));
-    }
+    nva.push_back(http2::make_nv_ls("x-forwarded-proto", scheme));
   }
 
-  auto via = downstream_->get_norm_request_header("via");
+  std::string via_value;
+  auto via = downstream_->get_request_header(http2::HD_VIA);
   if (get_config()->no_via) {
-    if (via != end_headers) {
+    if (via) {
       nva.push_back(http2::make_nv_ls("via", (*via).value));
     }
   } else {
-    if (via != end_headers) {
+    if (via) {
       via_value = (*via).value;
       via_value += ", ";
     }
@@ -426,22 +368,31 @@ int Http2DownstreamConnection::push_request_headers() {
     nva.push_back(http2::make_nv_ls("via", via_value));
   }
 
+  auto te = downstream_->get_request_header(http2::HD_TE);
+  // HTTP/1 upstream request can contain keyword other than
+  // "trailers".  We just forward "trailers".
+  // TODO more strict handling required here.
+  if (te && util::strifind(te->value.c_str(), "trailers")) {
+    nva.push_back(http2::make_nv_ll("te", "trailers"));
+  }
+
+  for (auto &p : get_config()->add_request_headers) {
+    nva.push_back(http2::make_nv(p.first, p.second));
+  }
+
   if (LOG_ENABLED(INFO)) {
     std::stringstream ss;
     for (auto &nv : nva) {
-      ss << TTY_HTTP_HD;
-      ss.write(reinterpret_cast<const char *>(nv.name), nv.namelen);
-      ss << TTY_RST << ": ";
-      ss.write(reinterpret_cast<const char *>(nv.value), nv.valuelen);
-      ss << "\n";
+      ss << TTY_HTTP_HD << nv.name << TTY_RST << ": " << nv.value << "\n";
     }
     DCLOG(INFO, this) << "HTTP request headers\n" << ss.str();
   }
 
   auto content_length =
-      downstream_->get_norm_request_header("content-length") != end_headers;
+      downstream_->get_request_header(http2::HD_CONTENT_LENGTH);
+  // TODO check content-length: 0 case
 
-  if (downstream_->get_request_method() == "CONNECT" || chunked_encoding ||
+  if (downstream_->get_request_method() == HTTP_CONNECT || chunked_encoding ||
       content_length || downstream_->get_request_http2_expect_body()) {
     // Request-body is expected.
     nghttp2_data_provider data_prd;
@@ -460,17 +411,15 @@ int Http2DownstreamConnection::push_request_headers() {
 
   downstream_->reset_downstream_wtimer();
 
-  http2session_->notify();
+  http2session_->signal_write();
   return 0;
 }
 
 int Http2DownstreamConnection::push_upload_data_chunk(const uint8_t *data,
                                                       size_t datalen) {
-  int rv = evbuffer_add(request_body_buf_, data, datalen);
-  if (rv != 0) {
-    DCLOG(FATAL, this) << "evbuffer_add() failed";
-    return -1;
-  }
+  int rv;
+  auto output = downstream_->get_request_buf();
+  output->append(data, datalen);
   if (downstream_->get_downstream_stream_id() != -1) {
     rv = http2session_->resume_data(this);
     if (rv != 0) {
@@ -479,7 +428,7 @@ int Http2DownstreamConnection::push_upload_data_chunk(const uint8_t *data,
 
     downstream_->ensure_downstream_wtimer();
 
-    http2session_->notify();
+    http2session_->signal_write();
   }
   return 0;
 }
@@ -494,7 +443,7 @@ int Http2DownstreamConnection::end_upload_data() {
 
     downstream_->ensure_downstream_wtimer();
 
-    http2session_->notify();
+    http2session_->signal_write();
   }
   return 0;
 }
@@ -524,7 +473,7 @@ int Http2DownstreamConnection::resume_read(IOCtrlReason reason,
 
     downstream_->dec_response_datalen(consumed);
 
-    http2session_->notify();
+    http2session_->signal_write();
   }
 
   return 0;
@@ -534,10 +483,6 @@ int Http2DownstreamConnection::on_read() { return 0; }
 
 int Http2DownstreamConnection::on_write() { return 0; }
 
-evbuffer *Http2DownstreamConnection::get_request_body_buf() const {
-  return request_body_buf_;
-}
-
 void Http2DownstreamConnection::attach_stream_data(StreamData *sd) {
   // It is possible sd->dconn is not NULL. sd is detached when
   // on_stream_close_callback. Before that, after MSG_COMPLETE is set
@@ -555,18 +500,8 @@ StreamData *Http2DownstreamConnection::detach_stream_data() {
     sd_ = nullptr;
     sd->dconn = nullptr;
     return sd;
-  } else {
-    return nullptr;
-  }
-}
-
-bool Http2DownstreamConnection::get_output_buffer_full() {
-  if (request_body_buf_) {
-    return evbuffer_get_length(request_body_buf_) >=
-           Http2Session::OUTBUF_MAX_THRES;
-  } else {
-    return false;
   }
+  return nullptr;
 }
 
 int Http2DownstreamConnection::on_priority_change(int32_t pri) {
@@ -583,7 +518,7 @@ int Http2DownstreamConnection::on_priority_change(int32_t pri) {
     DLOG(FATAL, this) << "nghttp2_submit_priority() failed";
     return -1;
   }
-  http2session_->notify();
+  http2session_->signal_write();
   return 0;
 }
 
@@ -595,4 +530,10 @@ int Http2DownstreamConnection::on_timeout() {
   return submit_rst_stream(downstream_, NGHTTP2_NO_ERROR);
 }
 
+size_t Http2DownstreamConnection::get_group() const {
+  // HTTP/2 backend connections are managed by Http2Session object,
+  // and it stores group index.
+  return http2session_->get_group();
+}
+
 } // namespace shrpx
