@@ -24,13 +24,30 @@
  */
 #include "util.h"
 
+#ifdef HAVE_TIME_H
 #include <time.h>
+#endif // HAVE_TIME_H
 #include <sys/types.h>
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
+#endif // HAVE_SYS_SOCKET_H
+#ifdef HAVE_NETDB_H
 #include <netdb.h>
+#endif // HAVE_NETDB_H
 #include <sys/stat.h>
+#ifdef HAVE_FCNTL_H
 #include <fcntl.h>
+#endif // HAVE_FCNTL_H
+#ifdef HAVE_NETINET_IN_H
+#include <netinet/in.h>
+#endif // HAVE_NETINET_IN_H
+#include <netinet/tcp.h>
+#ifdef HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif // HAVE_ARPA_INET_H
 
+#include <cmath>
+#include <cerrno>
 #include <cassert>
 #include <cstdio>
 #include <cstring>
@@ -39,6 +56,7 @@
 #include <nghttp2/nghttp2.h>
 
 #include "timegm.h"
+#include "template.h"
 
 namespace nghttp2 {
 
@@ -48,22 +66,19 @@ const char DEFAULT_STRIP_CHARSET[] = "\r\n\t ";
 
 const char UPPER_XDIGITS[] = "0123456789ABCDEF";
 
-bool isAlpha(const char c) {
-  return ('A' <= c && c <= 'Z') || ('a' <= c && c <= 'z');
-}
-
-bool isDigit(const char c) { return '0' <= c && c <= '9'; }
-
-bool isHexDigit(const char c) {
-  return isDigit(c) || ('A' <= c && c <= 'F') || ('a' <= c && c <= 'f');
-}
-
 bool inRFC3986UnreservedChars(const char c) {
   static const char unreserved[] = {'-', '.', '_', '~'};
   return isAlpha(c) || isDigit(c) ||
          std::find(&unreserved[0], &unreserved[4], c) != &unreserved[4];
 }
 
+bool in_rfc3986_sub_delims(const char c) {
+  static const char sub_delims[] = {'!', '$', '&', '\'', '(', ')',
+                                    '*', '+', ',', ';',  '='};
+  return std::find(std::begin(sub_delims), std::end(sub_delims), c) !=
+         std::end(sub_delims);
+}
+
 std::string percentEncode(const unsigned char *target, size_t len) {
   std::string dest;
   for (size_t i = 0; i < len; ++i) {
@@ -85,6 +100,21 @@ std::string percentEncode(const std::string &target) {
                        target.size());
 }
 
+std::string percent_encode_path(const std::string &s) {
+  std::string dest;
+  for (auto c : s) {
+    if (inRFC3986UnreservedChars(c) || in_rfc3986_sub_delims(c) || c == '/') {
+      dest += c;
+      continue;
+    }
+
+    dest += "%";
+    dest += UPPER_XDIGITS[(c >> 4) & 0x0f];
+    dest += UPPER_XDIGITS[(c & 0x0f)];
+  }
+  return dest;
+}
+
 bool in_token(char c) {
   static const char extra[] = {'!', '#', '$', '%', '&', '\'', '*', '+',
                                '-', '.', '^', '_', '`', '|',  '~'};
@@ -94,6 +124,12 @@ bool in_token(char c) {
              &extra[sizeof(extra)];
 }
 
+bool in_attr_char(char c) {
+  static const char bad[] = {'*', '\'', '%'};
+  return util::in_token(c) &&
+         std::find(std::begin(bad), std::end(bad) - 1, c) == std::end(bad) - 1;
+}
+
 std::string percent_encode_token(const std::string &target) {
   auto len = target.size();
   std::string dest;
@@ -112,24 +148,17 @@ std::string percent_encode_token(const std::string &target) {
   return dest;
 }
 
-std::string percentDecode(std::string::const_iterator first,
-                          std::string::const_iterator last) {
-  std::string result;
-  for (; first != last; ++first) {
-    if (*first == '%') {
-      if (first + 1 != last && first + 2 != last && isHexDigit(*(first + 1)) &&
-          isHexDigit(*(first + 2))) {
-        std::string numstr(first + 1, first + 3);
-        result += strtol(numstr.c_str(), 0, 16);
-        first += 2;
-      } else {
-        result += *first;
-      }
-    } else {
-      result += *first;
-    }
+uint32_t hex_to_uint(char c) {
+  if (c <= '9') {
+    return c - '0';
   }
-  return result;
+  if (c <= 'Z') {
+    return c - 'A' + 10;
+  }
+  if (c <= 'z') {
+    return c - 'a' + 10;
+  }
+  return c;
 }
 
 std::string quote_string(const std::string &target) {
@@ -187,13 +216,13 @@ std::string http_date(time_t t) {
   auto p = std::begin(res);
 
   auto s = DAY_OF_WEEK[tms.tm_wday];
-  p = std::copy(s, s + 3, p);
+  p = std::copy_n(s, 3, p);
   *p++ = ',';
   *p++ = ' ';
   p = cpydig(p, tms.tm_mday, 2);
   *p++ = ' ';
   s = MONTH[tms.tm_mon];
-  p = std::copy(s, s + 3, p);
+  p = std::copy_n(s, 3, p);
   *p++ = ' ';
   p = cpydig(p, tms.tm_year + 1900, 4);
   *p++ = ' ';
@@ -203,7 +232,7 @@ std::string http_date(time_t t) {
   *p++ = ':';
   p = cpydig(p, tms.tm_sec, 2);
   s = " GMT";
-  p = std::copy(s, s + 4, p);
+  p = std::copy_n(s, 4, p);
 
   return res;
 }
@@ -215,7 +244,6 @@ std::string common_log_date(time_t t) {
     return "";
   }
 
-#ifdef HAVE_STRUCT_TM_TM_GMTOFF
   // Format data like this:
   // 03/Jul/2014:00:19:38 +0900
   std::string res;
@@ -226,7 +254,7 @@ std::string common_log_date(time_t t) {
   p = cpydig(p, tms.tm_mday, 2);
   *p++ = '/';
   auto s = MONTH[tms.tm_mon];
-  p = std::copy(s, s + 3, p);
+  p = std::copy_n(s, 3, p);
   *p++ = '/';
   p = cpydig(p, tms.tm_year + 1900, 4);
   *p++ = ':';
@@ -237,7 +265,11 @@ std::string common_log_date(time_t t) {
   p = cpydig(p, tms.tm_sec, 2);
   *p++ = ' ';
 
+#ifdef HAVE_STRUCT_TM_TM_GMTOFF
   auto gmtoff = tms.tm_gmtoff;
+#else  // !HAVE_STRUCT_TM_TM_GMTOFF
+  auto gmtoff = nghttp2_timegm(&tms) - t;
+#endif // !HAVE_STRUCT_TM_TM_GMTOFF
   if (gmtoff >= 0) {
     *p++ = '+';
   } else {
@@ -249,13 +281,6 @@ std::string common_log_date(time_t t) {
   p = cpydig(p, (gmtoff % 3600) / 60, 2);
 
   return res;
-#else  // !HAVE_STRUCT_TM_TM_GMTOFF
-  char buf[32];
-
-  strftime(buf, sizeof(buf), "%d/%b/%Y:%T %z", &tms);
-
-  return buf;
-#endif // !HAVE_STRUCT_TM_TM_GMTOFF
 }
 
 std::string iso8601_date(int64_t ms) {
@@ -266,7 +291,6 @@ std::string iso8601_date(int64_t ms) {
     return "";
   }
 
-#ifdef HAVE_STRUCT_TM_TM_GMTOFF
   // Format data like this:
   // 2014-11-15T12:58:24.741Z
   // 2014-11-15T12:58:24.741+09:00
@@ -289,7 +313,11 @@ std::string iso8601_date(int64_t ms) {
   *p++ = '.';
   p = cpydig(p, ms % 1000, 3);
 
+#ifdef HAVE_STRUCT_TM_TM_GMTOFF
   auto gmtoff = tms.tm_gmtoff;
+#else  // !HAVE_STRUCT_TM_TM_GMTOFF
+  auto gmtoff = nghttp2_timegm(&tms) - sec;
+#endif // !HAVE_STRUCT_TM_TM_GMTOFF
   if (gmtoff == 0) {
     *p++ = 'Z';
   } else {
@@ -307,47 +335,15 @@ std::string iso8601_date(int64_t ms) {
   res.resize(p - std::begin(res));
 
   return res;
-#else  // !HAVE_STRUCT_TM_TM_GMTOFF
-  char buf[128];
-
-  auto nwrite = strftime(buf, sizeof(buf), "%Y-%m-%dT%H:%M:%S", &tms);
-  nwrite += snprintf(&buf[nwrite], sizeof(buf) - nwrite, ".%03d",
-                     static_cast<int>(ms % 1000));
-  auto nzone = strftime(&buf[nwrite], sizeof(buf) - nwrite, "%z", &tms);
-
-  // %z of strftime writes +hhmm or -hhmm not Z, +hh:mm or -hh:mm.  Do
-  // %nothing if nzone is not 5.  we don't know how to cope with this.
-  if (nzone == 5) {
-    if (memcmp(&buf[nwrite], "+0000", 5) == 0) {
-      // 0000 should be Z
-      memcpy(&buf[nwrite], "Z", 2);
-    } else {
-      // Move mm part to right by 1 including terminal \0
-      memmove(&buf[nwrite + 4], &buf[nwrite + 3], 3);
-      // Insert ':' between hh and mm
-      buf[nwrite + 3] = ':';
-    }
-  }
-  return buf;
-#endif // !HAVE_STRUCT_TM_TM_GMTOFF
 }
 
 time_t parse_http_date(const std::string &s) {
-  tm tm;
-  memset(&tm, 0, sizeof(tm));
+  tm tm{};
   char *r = strptime(s.c_str(), "%a, %d %b %Y %H:%M:%S GMT", &tm);
   if (r == 0) {
     return 0;
   }
-  return timegm(&tm);
-}
-
-bool startsWith(const std::string &a, const std::string &b) {
-  return startsWith(a.begin(), a.end(), b.begin(), b.end());
-}
-
-bool istartsWith(const std::string &a, const std::string &b) {
-  return istartsWith(a.begin(), a.end(), b.begin(), b.end());
+  return nghttp2_timegm_without_yday(&tm);
 }
 
 namespace {
@@ -365,26 +361,6 @@ bool istartsWith(const char *a, const char *b) {
   return !*b;
 }
 
-bool istartsWith(const char *a, size_t n, const char *b) {
-  return istartsWith(a, a + n, b, b + strlen(b));
-}
-
-bool endsWith(const std::string &a, const std::string &b) {
-  return endsWith(a.begin(), a.end(), b.begin(), b.end());
-}
-
-bool strieq(const std::string &a, const std::string &b) {
-  if (a.size() != b.size()) {
-    return false;
-  }
-  for (size_t i = 0; i < a.size(); ++i) {
-    if (lowcase(a[i]) != lowcase(b[i])) {
-      return false;
-    }
-  }
-  return true;
-}
-
 bool strieq(const char *a, const char *b) {
   if (!a || !b) {
     return false;
@@ -394,20 +370,6 @@ bool strieq(const char *a, const char *b) {
   return !*a && !*b;
 }
 
-bool strieq(const char *a, const uint8_t *b, size_t bn) {
-  if (!a || !b) {
-    return false;
-  }
-  const uint8_t *blast = b + bn;
-  for (; *a && b != blast && lowcase(*a) == lowcase(*b); ++a, ++b)
-    ;
-  return !*a && b == blast;
-}
-
-bool strieq(const char *a, const char *b, size_t bn) {
-  return strieq(a, reinterpret_cast<const uint8_t *>(b), bn);
-}
-
 int strcompare(const char *a, const uint8_t *b, size_t bn) {
   assert(a && b);
   const uint8_t *blast = b + bn;
@@ -468,56 +430,46 @@ std::string format_hex(const unsigned char *s, size_t len) {
 }
 
 void to_token68(std::string &base64str) {
-  for (auto i = std::begin(base64str); i != std::end(base64str); ++i) {
-    switch (*i) {
+  std::transform(std::begin(base64str), std::end(base64str),
+                 std::begin(base64str), [](char c) {
+    switch (c) {
     case '+':
-      *i = '-';
-      break;
+      return '-';
     case '/':
-      *i = '_';
-      break;
-    case '=':
-      base64str.erase(i, std::end(base64str));
-      return;
+      return '_';
+    default:
+      return c;
     }
-  }
-  return;
+  });
+  base64str.erase(std::find(std::begin(base64str), std::end(base64str), '='),
+                  std::end(base64str));
 }
 
 void to_base64(std::string &token68str) {
-  for (auto i = std::begin(token68str); i != std::end(token68str); ++i) {
-    switch (*i) {
+  std::transform(std::begin(token68str), std::end(token68str),
+                 std::begin(token68str), [](char c) {
+    switch (c) {
     case '-':
-      *i = '+';
-      break;
+      return '+';
     case '_':
-      *i = '/';
-      break;
+      return '/';
+    default:
+      return c;
     }
-  }
+  });
   if (token68str.size() & 0x3) {
     token68str.append(4 - (token68str.size() & 0x3), '=');
   }
   return;
 }
 
-void inp_strlower(std::string &s) {
-  for (auto i = std::begin(s); i != std::end(s); ++i) {
-    if ('A' <= *i && *i <= 'Z') {
-      *i = (*i) - 'A' + 'a';
-    }
-  }
-}
-
 namespace {
 // Calculates Damerau–Levenshtein distance between c-string a and b
 // with given costs.  swapcost, subcost, addcost and delcost are cost
 // to swap 2 adjacent characters, substitute characters, add character
 // and delete character respectively.
-int levenshtein(const char *a, const char *b, int swapcost, int subcost,
-                int addcost, int delcost) {
-  int alen = strlen(a);
-  int blen = strlen(b);
+int levenshtein(const char *a, int alen, const char *b, int blen, int swapcost,
+                int subcost, int addcost, int delcost) {
   auto dp = std::vector<std::vector<int>>(3, std::vector<int>(blen + 1));
   for (int i = 0; i <= blen; ++i) {
     dp[1][i] = i;
@@ -545,15 +497,21 @@ void show_candidates(const char *unkopt, option *options) {
   if (*unkopt == '\0') {
     return;
   }
+  auto unkoptend = unkopt;
+  for (; *unkoptend && *unkoptend != '='; ++unkoptend)
+    ;
+  auto unkoptlen = unkoptend - unkopt;
+  if (unkoptlen == 0) {
+    return;
+  }
   int prefix_match = 0;
-  auto unkoptlen = strlen(unkopt);
   auto cands = std::vector<std::pair<int, const char *>>();
   for (size_t i = 0; options[i].name != nullptr; ++i) {
     auto optnamelen = strlen(options[i].name);
     // Use cost 0 for prefix match
     if (istartsWith(options[i].name, options[i].name + optnamelen, unkopt,
                     unkopt + unkoptlen)) {
-      if (optnamelen == unkoptlen) {
+      if (optnamelen == static_cast<size_t>(unkoptlen)) {
         // Exact match, then we don't show any condidates.
         return;
       }
@@ -569,7 +527,8 @@ void show_candidates(const char *unkopt, option *options) {
       continue;
     }
     // cost values are borrowed from git, help.c.
-    int sim = levenshtein(unkopt, options[i].name, 0, 2, 1, 3);
+    int sim =
+        levenshtein(unkopt, unkoptlen, options[i].name, optnamelen, 0, 2, 1, 3);
     cands.emplace_back(sim, options[i].name);
   }
   if (prefix_match == 1 || cands.empty()) {
@@ -667,34 +626,60 @@ void write_uri_field(std::ostream &o, const char *uri, const http_parser_url &u,
 }
 
 bool numeric_host(const char *hostname) {
-  struct addrinfo hints;
-  struct addrinfo *res;
-  memset(&hints, 0, sizeof(hints));
-  hints.ai_family = AF_UNSPEC;
-  hints.ai_flags = AI_NUMERICHOST;
-  if (getaddrinfo(hostname, nullptr, &hints, &res)) {
-    return false;
-  }
-  freeaddrinfo(res);
-  return true;
+  return numeric_host(hostname, AF_INET) || numeric_host(hostname, AF_INET6);
 }
 
-int reopen_log_file(const char *path) {
-#if defined(__ANDROID__) || defined(ANDROID)
-  int fd;
+bool numeric_host(const char *hostname, int family) {
+  int rv;
+  std::array<uint8_t, sizeof(struct in6_addr)> dst;
 
-  if (strcmp("/proc/self/fd/1", path) == 0 ||
-      strcmp("/proc/self/fd/2", path) == 0) {
+  rv = inet_pton(family, hostname, dst.data());
 
-    // We will get permission denied error when O_APPEND is used for
-    // these paths.
-    fd =
-        open(path, O_WRONLY | O_CREAT | O_CLOEXEC, S_IRUSR | S_IWUSR | S_IRGRP);
-  } else {
-    fd = open(path, O_WRONLY | O_APPEND | O_CREAT | O_CLOEXEC,
-              S_IRUSR | S_IWUSR | S_IRGRP);
+  return rv == 1;
+}
+
+std::string numeric_name(const struct sockaddr *sa, socklen_t salen) {
+  std::array<char, NI_MAXHOST> host;
+  auto rv = getnameinfo(sa, salen, host.data(), host.size(), nullptr, 0,
+                        NI_NUMERICHOST);
+  if (rv != 0) {
+    return "unknown";
   }
-#elif defined O_CLOEXEC
+  return host.data();
+}
+
+static int STDERR_COPY = -1;
+static int STDOUT_COPY = -1;
+
+void store_original_fds() {
+  // consider dup'ing stdout too
+  STDERR_COPY = dup(STDERR_FILENO);
+  STDOUT_COPY = STDOUT_FILENO;
+  // no race here, since it is called early
+  make_socket_closeonexec(STDERR_COPY);
+}
+
+void restore_original_fds() { dup2(STDERR_COPY, STDERR_FILENO); }
+
+void close_log_file(int &fd) {
+  if (fd != STDERR_COPY && fd != STDOUT_COPY && fd != -1) {
+    close(fd);
+  }
+  fd = -1;
+}
+
+int open_log_file(const char *path) {
+
+  if (strcmp(path, "/dev/stdout") == 0 ||
+      strcmp(path, "/proc/self/fd/1") == 0) {
+    return STDOUT_COPY;
+  }
+
+  if (strcmp(path, "/dev/stderr") == 0 ||
+      strcmp(path, "/proc/self/fd/2") == 0) {
+    return STDERR_COPY;
+  }
+#if defined O_CLOEXEC
 
   auto fd = open(path, O_WRONLY | O_APPEND | O_CREAT | O_CLOEXEC,
                  S_IRUSR | S_IWUSR | S_IRGRP);
@@ -705,7 +690,7 @@ int reopen_log_file(const char *path) {
 
   // We get race condition if execve is called at the same time.
   if (fd != -1) {
-    fcntl(fd, F_SETFD, FD_CLOEXEC);
+    make_socket_closeonexec(fd);
   }
 
 #endif // !O_CLOEXEC
@@ -745,10 +730,16 @@ char *get_exec_path(int argc, char **const argv, const char *cwd) {
 
   if (argv0[0] == '/') {
     path = static_cast<char *>(malloc(len + 1));
+    if (path == nullptr) {
+      return nullptr;
+    }
     memcpy(path, argv0, len + 1);
   } else {
     auto cwdlen = strlen(cwd);
     path = static_cast<char *>(malloc(len + 1 + cwdlen + 1));
+    if (path == nullptr) {
+      return nullptr;
+    }
     memcpy(path, cwd, cwdlen);
     path[cwdlen] = '/';
     memcpy(path + cwdlen + 1, argv0, len + 1);
@@ -771,20 +762,461 @@ int64_t to_time64(const timeval &tv) {
 }
 
 bool check_h2_is_selected(const unsigned char *proto, size_t len) {
-  return streq(NGHTTP2_PROTO_VERSION_ID, NGHTTP2_PROTO_VERSION_ID_LEN, proto,
-               len);
+  return streq_l(NGHTTP2_PROTO_VERSION_ID, proto, len) ||
+         streq_l(NGHTTP2_H2_16, proto, len) ||
+         streq_l(NGHTTP2_H2_14, proto, len);
+}
+
+namespace {
+bool select_proto(const unsigned char **out, unsigned char *outlen,
+                  const unsigned char *in, unsigned int inlen, const char *key,
+                  unsigned int keylen) {
+  for (auto p = in, end = in + inlen; p + keylen <= end; p += *p + 1) {
+    if (std::equal(key, key + keylen, p)) {
+      *out = p + 1;
+      *outlen = *p;
+      return true;
+    }
+  }
+  return false;
+}
+} // namespace
+
+bool select_h2(const unsigned char **out, unsigned char *outlen,
+               const unsigned char *in, unsigned int inlen) {
+  return select_proto(out, outlen, in, inlen, NGHTTP2_PROTO_ALPN,
+                      str_size(NGHTTP2_PROTO_ALPN)) ||
+         select_proto(out, outlen, in, inlen, NGHTTP2_H2_16_ALPN,
+                      str_size(NGHTTP2_H2_16_ALPN)) ||
+         select_proto(out, outlen, in, inlen, NGHTTP2_H2_14_ALPN,
+                      str_size(NGHTTP2_H2_14_ALPN));
+}
+
+bool select_protocol(const unsigned char **out, unsigned char *outlen,
+                     const unsigned char *in, unsigned int inlen,
+                     std::vector<std::string> proto_list) {
+  for (const auto &proto : proto_list) {
+    if (select_proto(out, outlen, in, inlen, proto.c_str(),
+                     static_cast<unsigned int>(proto.size()))) {
+      return true;
+    }
+  }
+
+  return false;
 }
 
 std::vector<unsigned char> get_default_alpn() {
-  auto res = std::vector<unsigned char>(1 + NGHTTP2_PROTO_VERSION_ID_LEN);
-  auto p = res.data();
+  auto res = std::vector<unsigned char>(str_size(NGHTTP2_PROTO_ALPN) +
+                                        str_size(NGHTTP2_H2_16_ALPN) +
+                                        str_size(NGHTTP2_H2_14_ALPN));
+  auto p = std::begin(res);
 
-  *p++ = NGHTTP2_PROTO_VERSION_ID_LEN;
-  memcpy(p, NGHTTP2_PROTO_VERSION_ID, NGHTTP2_PROTO_VERSION_ID_LEN);
+  p = std::copy_n(NGHTTP2_PROTO_ALPN, str_size(NGHTTP2_PROTO_ALPN), p);
+  p = std::copy_n(NGHTTP2_H2_16_ALPN, str_size(NGHTTP2_H2_16_ALPN), p);
+  p = std::copy_n(NGHTTP2_H2_14_ALPN, str_size(NGHTTP2_H2_14_ALPN), p);
 
   return res;
 }
 
+std::vector<Range<const char *>> split_config_str_list(const char *s,
+                                                       char delim) {
+  size_t len = 1;
+  auto last = s + strlen(s);
+  for (const char *first = s, *d = nullptr;
+       (d = std::find(first, last, delim)) != last; ++len, first = d + 1)
+    ;
+
+  auto list = std::vector<Range<const char *>>(len);
+
+  len = 0;
+  for (auto first = s;; ++len) {
+    auto stop = std::find(first, last, delim);
+    list[len] = {first, stop};
+    if (stop == last) {
+      break;
+    }
+    first = stop + 1;
+  }
+  return list;
+}
+
+std::vector<std::string> parse_config_str_list(const char *s, char delim) {
+  auto ranges = split_config_str_list(s, delim);
+  auto res = std::vector<std::string>();
+  res.reserve(ranges.size());
+  for (const auto &range : ranges) {
+    res.emplace_back(range.first, range.second);
+  }
+  return res;
+}
+
+int make_socket_closeonexec(int fd) {
+  int flags;
+  int rv;
+  while ((flags = fcntl(fd, F_GETFD)) == -1 && errno == EINTR)
+    ;
+  while ((rv = fcntl(fd, F_SETFD, flags | FD_CLOEXEC)) == -1 && errno == EINTR)
+    ;
+  return rv;
+}
+
+int make_socket_nonblocking(int fd) {
+  int flags;
+  int rv;
+  while ((flags = fcntl(fd, F_GETFL, 0)) == -1 && errno == EINTR)
+    ;
+  while ((rv = fcntl(fd, F_SETFL, flags | O_NONBLOCK)) == -1 && errno == EINTR)
+    ;
+  return rv;
+}
+
+int make_socket_nodelay(int fd) {
+  int val = 1;
+  if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, reinterpret_cast<char *>(&val),
+                 sizeof(val)) == -1) {
+    return -1;
+  }
+  return 0;
+}
+
+int create_nonblock_socket(int family) {
+#ifdef SOCK_NONBLOCK
+  auto fd = socket(family, SOCK_STREAM | SOCK_NONBLOCK | SOCK_CLOEXEC, 0);
+
+  if (fd == -1) {
+    return -1;
+  }
+#else  // !SOCK_NONBLOCK
+  auto fd = socket(family, SOCK_STREAM, 0);
+
+  if (fd == -1) {
+    return -1;
+  }
+
+  make_socket_nonblocking(fd);
+  make_socket_closeonexec(fd);
+#endif // !SOCK_NONBLOCK
+
+  if (family == AF_INET || family == AF_INET6) {
+    make_socket_nodelay(fd);
+  }
+
+  return fd;
+}
+
+bool check_socket_connected(int fd) {
+  int error;
+  socklen_t len = sizeof(error);
+  if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len) == 0) {
+    if (error != 0) {
+      return false;
+    }
+  }
+  return true;
+}
+
+bool ipv6_numeric_addr(const char *host) {
+  uint8_t dst[16];
+  return inet_pton(AF_INET6, host, dst) == 1;
+}
+
+namespace {
+std::pair<int64_t, size_t> parse_uint_digits(const void *ss, size_t len) {
+  const uint8_t *s = static_cast<const uint8_t *>(ss);
+  int64_t n = 0;
+  size_t i;
+  if (len == 0) {
+    return {-1, 0};
+  }
+  constexpr int64_t max = std::numeric_limits<int64_t>::max();
+  for (i = 0; i < len; ++i) {
+    if ('0' <= s[i] && s[i] <= '9') {
+      if (n > max / 10) {
+        return {-1, 0};
+      }
+      n *= 10;
+      if (n > max - (s[i] - '0')) {
+        return {-1, 0};
+      }
+      n += s[i] - '0';
+      continue;
+    }
+    break;
+  }
+  if (i == 0) {
+    return {-1, 0};
+  }
+  return {n, i};
+}
+} // namespace
+
+int64_t parse_uint_with_unit(const char *s) {
+  int64_t n;
+  size_t i;
+  auto len = strlen(s);
+  std::tie(n, i) = parse_uint_digits(s, len);
+  if (n == -1) {
+    return -1;
+  }
+  if (i == len) {
+    return n;
+  }
+  if (i + 1 != len) {
+    return -1;
+  }
+  int mul = 1;
+  switch (s[i]) {
+  case 'K':
+  case 'k':
+    mul = 1 << 10;
+    break;
+  case 'M':
+  case 'm':
+    mul = 1 << 20;
+    break;
+  case 'G':
+  case 'g':
+    mul = 1 << 30;
+    break;
+  default:
+    return -1;
+  }
+  constexpr int64_t max = std::numeric_limits<int64_t>::max();
+  if (n > max / mul) {
+    return -1;
+  }
+  return n * mul;
+}
+
+int64_t parse_uint(const char *s) {
+  return parse_uint(reinterpret_cast<const uint8_t *>(s), strlen(s));
+}
+
+int64_t parse_uint(const std::string &s) {
+  return parse_uint(reinterpret_cast<const uint8_t *>(s.c_str()), s.size());
+}
+
+int64_t parse_uint(const uint8_t *s, size_t len) {
+  int64_t n;
+  size_t i;
+  std::tie(n, i) = parse_uint_digits(s, len);
+  if (n == -1 || i != len) {
+    return -1;
+  }
+  return n;
+}
+
+double parse_duration_with_unit(const char *s) {
+  constexpr auto max = std::numeric_limits<int64_t>::max();
+  int64_t n;
+  size_t i;
+  auto len = strlen(s);
+  std::tie(n, i) = parse_uint_digits(s, len);
+  if (n == -1) {
+    goto fail;
+  }
+  if (i == len) {
+    return static_cast<double>(n);
+  }
+  switch (s[i]) {
+  case 'S':
+  case 's':
+    // seconds
+    if (i + 1 != len) {
+      goto fail;
+    }
+    return static_cast<double>(n);
+  case 'M':
+  case 'm':
+    if (i + 1 == len) {
+      // minutes
+      if (n > max / 60) {
+        goto fail;
+      }
+      return static_cast<double>(n) * 60;
+    }
+
+    if (i + 2 != len || (s[i + 1] != 's' && s[i + 1] != 'S')) {
+      goto fail;
+    }
+    // milliseconds
+    return static_cast<double>(n) / 1000.;
+  case 'H':
+  case 'h':
+    // hours
+    if (i + 1 != len) {
+      goto fail;
+    }
+    if (n > max / 3600) {
+      goto fail;
+    }
+    return static_cast<double>(n) * 3600;
+  }
+fail:
+  return std::numeric_limits<double>::infinity();
+}
+
+std::string duration_str(double t) {
+  if (t == 0.) {
+    return "0";
+  }
+  auto frac = static_cast<int64_t>(t * 1000) % 1000;
+  if (frac > 0) {
+    return utos(static_cast<int64_t>(t * 1000)) + "ms";
+  }
+  auto v = static_cast<int64_t>(t);
+  if (v % 60) {
+    return utos(v) + "s";
+  }
+  v /= 60;
+  if (v % 60) {
+    return utos(v) + "m";
+  }
+  v /= 60;
+  return utos(v) + "h";
+}
+
+std::string format_duration(const std::chrono::microseconds &u) {
+  const char *unit = "us";
+  int d = 0;
+  auto t = u.count();
+  if (t >= 1000000) {
+    d = 1000000;
+    unit = "s";
+  } else if (t >= 1000) {
+    d = 1000;
+    unit = "ms";
+  } else {
+    return utos(t) + unit;
+  }
+  return dtos(static_cast<double>(t) / d) + unit;
+}
+
+std::string dtos(double n) {
+  auto f = utos(static_cast<int64_t>(round(100. * n)) % 100);
+  return utos(static_cast<int64_t>(n)) + "." + (f.size() == 1 ? "0" : "") + f;
+}
+
+std::string make_hostport(const char *host, uint16_t port) {
+  auto ipv6 = ipv6_numeric_addr(host);
+  std::string hostport;
+
+  if (ipv6) {
+    hostport += "[";
+  }
+
+  hostport += host;
+
+  if (ipv6) {
+    hostport += "]";
+  }
+
+  if (port != 80 && port != 443) {
+    hostport += ":";
+    hostport += utos(port);
+  }
+
+  return hostport;
+}
+
+namespace {
+void hexdump8(FILE *out, const uint8_t *first, const uint8_t *last) {
+  auto stop = std::min(first + 8, last);
+  for (auto k = first; k != stop; ++k) {
+    fprintf(out, "%02x ", *k);
+  }
+  // each byte needs 3 spaces (2 hex value and space)
+  for (; stop != first + 8; ++stop) {
+    fputs("   ", out);
+  }
+  // we have extra space after 8 bytes
+  fputc(' ', out);
+}
+} // namespace
+
+void hexdump(FILE *out, const uint8_t *src, size_t len) {
+  if (len == 0) {
+    return;
+  }
+  size_t buflen = 0;
+  auto repeated = false;
+  std::array<uint8_t, 16> buf{};
+  auto end = src + len;
+  auto i = src;
+  for (;;) {
+    auto nextlen =
+        std::min(static_cast<size_t>(16), static_cast<size_t>(end - i));
+    if (nextlen == buflen &&
+        std::equal(std::begin(buf), std::begin(buf) + buflen, i)) {
+      // as long as adjacent 16 bytes block are the same, we just
+      // print single '*'.
+      if (!repeated) {
+        repeated = true;
+        fputs("*\n", out);
+      }
+      i += nextlen;
+      continue;
+    }
+    repeated = false;
+    fprintf(out, "%08lx", static_cast<unsigned long>(i - src));
+    if (i == end) {
+      fputc('\n', out);
+      break;
+    }
+    fputs("  ", out);
+    hexdump8(out, i, end);
+    hexdump8(out, i + 8, std::max(i + 8, end));
+    fputc('|', out);
+    auto stop = std::min(i + 16, end);
+    buflen = stop - i;
+    auto p = buf.data();
+    for (; i != stop; ++i) {
+      *p++ = *i;
+      if (0x20 <= *i && *i <= 0x7e) {
+        fputc(*i, out);
+      } else {
+        fputc('.', out);
+      }
+    }
+    fputs("|\n", out);
+  }
+}
+
+void put_uint16be(uint8_t *buf, uint16_t n) {
+  uint16_t x = htons(n);
+  memcpy(buf, &x, sizeof(uint16_t));
+}
+
+void put_uint32be(uint8_t *buf, uint32_t n) {
+  uint32_t x = htonl(n);
+  memcpy(buf, &x, sizeof(uint32_t));
+}
+
+uint16_t get_uint16(const uint8_t *data) {
+  uint16_t n;
+  memcpy(&n, data, sizeof(uint16_t));
+  return ntohs(n);
+}
+
+uint32_t get_uint32(const uint8_t *data) {
+  uint32_t n;
+  memcpy(&n, data, sizeof(uint32_t));
+  return ntohl(n);
+}
+
+uint64_t get_uint64(const uint8_t *data) {
+  uint64_t n = 0;
+  n += static_cast<uint64_t>(data[0]) << 56;
+  n += static_cast<uint64_t>(data[1]) << 48;
+  n += static_cast<uint64_t>(data[2]) << 40;
+  n += static_cast<uint64_t>(data[3]) << 32;
+  n += static_cast<uint64_t>(data[4]) << 24;
+  n += data[5] << 16;
+  n += data[6] << 8;
+  n += data[7];
+  return n;
+}
+
 } // namespace util
 
 } // namespace nghttp2
